#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/time.h>
#include <xtalk/debug.h>
#include <xpanel/pcptalk.h>
#include <xpanel/kbd_glib.h>

#define	DBG_MASK	0x04

#define	KBD_POLL_MS	200	/* Keyboard poll interval (msec) */

GSource		*kbd_source;

const char *idle_state_name(enum kbd_idle_state state)
{
#define	ENTRY(x)	[KBD_IDLE_ ## x] = #x
	static const char *names[] = {
		ENTRY(NONE),
		ENTRY(START),
		ENTRY(KEEP_GOING),
		ENTRY(STOPPED),
		ENTRY(STOP),
	};
#undef ENTRY
	if (state >= sizeof(names)/sizeof(names[0]))
		return NULL;
	return names[state];
}

void kbd_idle_set_callback(struct xpanel_kbd *kbd, kbd_callback_t callback, void *user_data)
{
	DBG("%p\n", callback);
	kbd->kbd_idle_callback = callback;
	kbd->kbd_idle_data = user_data;
	kbd_idle_set_timeout(kbd, 10);	/* Default */
}

static void kbd_idle_set_state(struct xpanel_kbd *kbd, enum kbd_idle_state new_idle_state)
{
	if (kbd->idle_state == new_idle_state)
		return;
	DBG("%s: %s -> %s\n", __func__,
		idle_state_name(kbd->idle_state),
		idle_state_name(new_idle_state));
	kbd->idle_state = new_idle_state;
	if (kbd->kbd_idle_callback)
		kbd->kbd_idle_callback(kbd, kbd->kbd_idle_data);
}

void kbd_idle_set_timeout(struct xpanel_kbd *kbd, uint32_t seconds)
{
	DBG("%d seconds\n", seconds);
	kbd->kbd_idle_seconds = seconds;
}

static void kbd_state_is_idle(struct xpanel_kbd *kbd)
{
	switch (kbd->idle_state) {
		case KBD_IDLE_NONE:
		case KBD_IDLE_STOP:
		case KBD_IDLE_STOPPED:
			/* 1-shot */
			kbd_idle_set_state(kbd, KBD_IDLE_START);
			break;
		case KBD_IDLE_START:
			/* 1-shot */
			kbd_idle_set_state(kbd, KBD_IDLE_KEEP_GOING);
			break;
		case KBD_IDLE_KEEP_GOING:
			break;	/* same, same */
	}
}

static void kbd_state_is_not_idle(struct xpanel_kbd *kbd)
{
	switch (kbd->idle_state) {
		case KBD_IDLE_NONE:
		case KBD_IDLE_START:
		case KBD_IDLE_KEEP_GOING:
			/* 1-shot */
			kbd_idle_set_state(kbd, KBD_IDLE_STOP);
			break;
		case KBD_IDLE_STOP:
			/* 1-shot */
			kbd_idle_set_state(kbd, KBD_IDLE_STOPPED);
			break;
		case KBD_IDLE_STOPPED:
			break;	/* same, same */
	}
}

void kbd_idle_reset(struct xpanel_kbd *kbd)
{
	kbd_state_is_not_idle(kbd);
}

static gboolean kbd_prepare(GSource *source, gint *timeout)
{
	struct xpanel_kbd	*kbd = (struct xpanel_kbd *)source;

	DBG("kbd prepare(%d keys unprocessed)\n", kbd->numkeys);
	*timeout = KBD_POLL_MS;
	gettimeofday(&kbd->tv, NULL);
	return (kbd->numkeys) ? TRUE : FALSE;
}
 
static gboolean kbd_check(GSource *source)
{
	struct xpanel_kbd	*kbd = (struct xpanel_kbd *)source;
	struct timeval		now;
	long			usec_diff;
	int			numkeys;
	int			idle_seconds;

	gettimeofday(&now, NULL);
	usec_diff = (now.tv_sec - kbd->tv.tv_sec) * 1000 * 1000 + (now.tv_usec - kbd->tv.tv_usec);
	DBG("kbd check(%ld msec from prepare)\n", usec_diff/1000);
	idle_seconds = now.tv_sec - kbd->lastkey.tv_sec;
	// Turn off the backlight if needed.
	if( pcp_getbacklightstatus(kbd->pcpdev) && idle_seconds > PCP_BACKLIGHT_TO)
		pcp_backlight(kbd->pcpdev, 0);
	if (idle_seconds > kbd->kbd_idle_seconds) {
		/* kbd is idle */
		kbd_state_is_idle(kbd);
	}
	memset(kbd->keybuf, 0, sizeof(kbd->keybuf));
	if((numkeys = pcp_keyget(kbd->pcpdev, kbd->keybuf)) < 0) {
		ERR("pcp_keyget() failed\n");
	} else {
		kbd->numkeys = numkeys;
	}
	return (numkeys > 0) ? TRUE : FALSE;
}

static gboolean kbd_callback(gpointer user_data)
{
	struct xpanel_kbd	*kbd = (struct xpanel_kbd *)user_data;

	if(!kbd) {
		ERR("kbd callback, but no user_data\n");
		return TRUE;
	}
	if(kbd->numkeys == 0)
		return TRUE;
	kbd_idle_reset(kbd);
	
	// Save the timestamp of the key pressing. It is needed for the
	// backlight control.
	gettimeofday(&(kbd->lastkey), NULL);
	// Switch the backlight on if it is currently off
	if( ! pcp_getbacklightstatus(kbd->pcpdev))
		pcp_backlight(kbd->pcpdev, 1);	
#if 0
	INFO("kbd callback (%zd keys)\n", kbd->numkeys);
	for(i = 0; kbd->keybuf[i]; i++) {
		INFO("[%d]: Key(%d) %s\n", i, kbd->keybuf[i],
			keycode_name(kbd->keybuf[i]));
	}
#endif
	assert(kbd->notifykeys);
	if(kbd->notifykeys(kbd) == 0)
		return TRUE;
	return FALSE;
}

static gboolean kbd_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
	if(callback) {
		//INFO("Calling kbd callback\n");
		return callback(user_data);
	}
	//INFO("No kbd callback\n");
	return FALSE;
}


GSourceFuncs	kbd_funcs = {
	.prepare	= kbd_prepare,
	.check		= kbd_check,
	.dispatch	= kbd_dispatch
};

struct xpanel_kbd *kbd_glib_register(struct pcp_device *pcpdev, GMainContext *mainctx, kbd_notifykeys_t notifykeys)
{
	struct xpanel_kbd	*kbd;

	INFO("kbd_glib_register\n");
	if(!notifykeys) {
		ERR("%s: Missing a notifykeys() function parameter\n", __FUNCTION__);
		return NULL;
	}
	if((kbd_source = g_source_new(&kbd_funcs, sizeof(struct xpanel_kbd))) == NULL) {
		ERR("kbd source\n");
		return NULL;
	}
	kbd = (struct xpanel_kbd *)kbd_source;
	kbd->pcpdev = pcpdev;
	kbd->notifykeys = notifykeys;
	gettimeofday(&(kbd->lastkey), NULL); //Ensure that the backligt will be on during first feew minutes.
	g_source_attach(kbd_source, mainctx);
	g_source_set_callback(kbd_source, kbd_callback, kbd_source, NULL);
	return kbd;
}

void kbd_glib_destroy(struct xpanel_kbd *kbd)
{
	if(kbd) {
		memset(kbd, 0, sizeof(kbd));
		g_free(kbd);
	}
}

void kbd_flush(struct xpanel_kbd *kbd)
{
	uint8_t	buf[BUFSIZ];

	DBG("\n");
	pcp_keyget(kbd->pcpdev, buf);
}
