/*
 * Written by Oron Peled <oron@actcom.co.il>
 * Copyright (C) 2009, Xorcom
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <xtalk/debug.h>
#include <xtalk/proto_sync.h>
#include <xpanel/pcptalk.h>
#include <xpanel/firmware_defs.h>

static const char rcsid[] = "$Id$";

#define	DBG_MASK	0x02

#define	BIT(i)		(1 << (i))

#if	PCP_MAX_KEYBUF != MAX_KEYBUF
#error PCP_MAX_KEYBUF is different than MAX_KEYBUF
#endif

#if	PCP_MAX_FANS != FANS_NUM
#error PCP_MAX_FANS is different than FANS_NUM
#endif

/*
 * Interfaces for XBee remote
 */
struct zigbee_remote {
	int	lamp_state;
};

/*
 * It's just an XTALK device with a USB transport
 */
struct pcp_device {
	struct xtalk_base	*xtalk_base;
	struct xtalk_sync	*xtalk_sync;
	uint8_t			control_byte;	/* current state */
	uint8_t			panel_proto_version;
	/*
	 * Legal Values: 0...FAN_MAX_SPEED
	 * Unset: FAN_SPEED_UNSET
	 */
#define	FAN_SPEED_UNSET	0xFF
	uint8_t			fan_speed[FANS_NUM];
	uint8_t			backlight_status;
	uint8_t			psu_status;
	struct zigbee_remote	zigbee_remote;
};

CMD_DEF(PCP, LCD_SENDR,	/* Register */
	uint8_t	data;
	);

CMD_DEF(PCP, LCD_SENDD,	/* Display */
	char	ascii;
	);

CMD_DEF(PCP, LED_SEND,
	uint8_t	bitmask;
	);

CMD_DEF(PCP, KEY_GET,
	);

CMD_DEF(PCP, KEY_GET_REPLY,
	uint16_t	keys[0];	/* maximum MAX_KEYBUF */
	);

CMD_DEF(PCP, CONTROL,
	uint8_t	bitmask;
	);

CMD_DEF(PCP, RESET);

CMD_DEF(PCP, FAN_SET,
	uint8_t	fan_number;
	uint8_t	fan_speed;
	);

CMD_DEF(PCP, FAN_GET,
	uint8_t	fan_number;
	);

CMD_DEF(PCP, FAN_GET_REPLY,
	uint8_t	fan_number;
	uint8_t	fan_speed;
	);

CMD_DEF(PCP, PSU_GET,
	);

CMD_DEF(PCP, PSU_GET_REPLY,
	uint8_t	psu_status;
	uint8_t	psu_reserved;
	);

CMD_DEF(PCP, XBEE_SEND,
// off: 71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 31 05 6E
// on:  71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 31 04 6F
	uint8_t data[20];
	);

union XTALK_PDATA(PCP) {
	MEMBER(PCP, LCD_SENDR);
	MEMBER(PCP, LCD_SENDD);
	MEMBER(PCP, CONTROL);
	MEMBER(PCP, LED_SEND);
	MEMBER(PCP, KEY_GET);
	MEMBER(PCP, KEY_GET_REPLY);
	MEMBER(PCP, RESET);
	MEMBER(PCP, FAN_SET);
	MEMBER(PCP, FAN_GET);
	MEMBER(PCP, FAN_GET_REPLY);
	MEMBER(PCP, PSU_GET);
	MEMBER(PCP, PSU_GET_REPLY);
	MEMBER(PCP, XBEE_SEND);
} PACKED members;

struct xtalk_protocol	pcptalk_base = {
	.name	= "PCP",
	.proto_version = 0x11,
	.commands = {
		CMD_SEND(PCP, LCD_SENDR),
		CMD_SEND(PCP, LCD_SENDD),
		CMD_SEND(PCP, CONTROL),
		CMD_SEND(PCP, LED_SEND),
		CMD_SEND(PCP, KEY_GET),
		CMD_RECV(PCP, KEY_GET_REPLY),
		CMD_SEND(PCP, RESET),
		CMD_SEND(PCP, FAN_SET),
		CMD_SEND(PCP, FAN_GET),
		CMD_RECV(PCP, FAN_GET_REPLY),
		CMD_SEND(PCP, PSU_GET),
		CMD_RECV(PCP, PSU_GET_REPLY),
		CMD_SEND(PCP, XBEE_SEND),
	},
	.ack_statuses = {
	}
};

/*
 * Convenience wrappers
 */
static struct xtalk_command *pcp_new_command(
	struct pcp_device *pcpdev,
	uint8_t op,
	uint16_t extra_data)
{
	struct xtalk_command *cmd;

	if((cmd = new_command(pcpdev->xtalk_base, op, extra_data)) == NULL) {
		ERR("new_command failed\n");
		return NULL;
	}
	return cmd;
}

static int pcp_process_command(
	struct pcp_device *pcpdev,
	struct xtalk_command *cmd,
	struct xtalk_command **reply_ref,
	uint16_t *sequence_number)
{
	int ret;

	ret = process_command(pcpdev->xtalk_sync,
			cmd, reply_ref, sequence_number);
	if(ret < 0) {
		ERR("process_command failed(seq=%d): %d\n",
			*sequence_number, ret);
		return ret;
	}
	return ret;
}

int pcp_reset(struct pcp_device *pcpdev)
{
	struct xtalk_command	*cmd;
	uint16_t		tx_seq;
	int			ret;

	DBG("reset\n");
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_RESET, 0)) == NULL)
		return -ENOMEM;
	ret = pcp_process_command(pcpdev, cmd, NULL, &tx_seq);
	if(ret < 0) {
		ERR("process_command failed: %d\n", ret);
		return ret;
	}
	return 0;
}

uint8_t pcp_get_panel_proto_version(struct pcp_device *pcpdev)
{
	return pcpdev->panel_proto_version;
}

static int pcp_lcd_op(struct pcp_device *pcpdev, uint16_t data)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	//DBG("data=0x%04X\n", data);
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_LCD_SENDR, 0)) == NULL) {
		ERR("new_command failed\n");
		return -ENOMEM;
	}
	CMD_FIELD(cmd, PCP, LCD_SENDR, data) = data;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0) {
		ERR("process_command failed: %d\n", ret);
		goto out;;
	}
out:
	free_command(reply);
	return ret;
}

int pcp_lcd_clear_display(struct pcp_device *pcpdev)
{
	int	ret;

	DBG("\n");
	ret = pcp_lcd_op(pcpdev, BIT(0));
	usleep(1640);
	return ret;
}

int pcp_lcd_cursor_home(struct pcp_device *pcpdev)
{
	int	ret;

	DBG("\n");
	ret = pcp_lcd_op(pcpdev, BIT(1));
	usleep(1640);
	return ret;
}

int pcp_lcd_entry_mode(struct pcp_device *pcpdev, int i_d, int s)
{
	int		ret;
	uint16_t	mask = BIT(2);

	DBG("cursor-incr/decr=%d auto-shift=%d\n", i_d, s);
	if(i_d)	/* Increment/Decrement cursor position */
		mask |= BIT(1);
	if(s)	/* Automatic display shift */
		mask |= BIT(0);
	ret = pcp_lcd_op(pcpdev, mask);
	usleep(40);
	return ret;
}

int pcp_lcd_display_control(struct pcp_device *pcpdev, int d, int c, int b)
{
	int		ret;
	uint16_t	mask = BIT(3);

	DBG("display=%s cursor=%s blink=%s\n",
		(d) ? "on" : "off",
		(c) ? "on" : "off",
		(b) ? "on" : "off");
	if(d)	/* Display on/off */
		mask |= BIT(2);
	if(c)	/* Cursor on/off */
		mask |= BIT(1);
	if(b)	/* Cursor blink on/off */
		mask |= BIT(0);
	ret = pcp_lcd_op(pcpdev, mask);
	usleep(40);
	return ret;
}

int pcp_lcd_display_shift(struct pcp_device *pcpdev, int s_c, int r_l)
{
	int		ret;
	uint16_t	mask = BIT(4);

	DBG("shift/move=%d right/left=%d\n", s_c, r_l);
	if(s_c)	/* Display shift or Cursor move */
		mask |= BIT(3);
	if(r_l)	/* Shift right/left */
		mask |= BIT(2);
	ret = pcp_lcd_op(pcpdev, mask);
	usleep(40);
	return ret;
}

int pcp_lcd_function_set(struct pcp_device *pcpdev, int d_l, int n, int f)
{
	int		ret;
	uint16_t	mask = BIT(5);

	DBG("8/4bit=%d 2/1lines=%d 5/10font=%d\n", d_l, n, f);
	if(d_l)	/* 8/4 bit interface   */
		mask |= BIT(4);
	if(n)	/* 2/1 display lines   */
		mask |= BIT(3);
	if(f)	/* 5/10 font, 5/7 font */
		mask |= BIT(2);
	ret = pcp_lcd_op(pcpdev, mask);
	usleep(40);
	return ret;
}

int pcp_lcd_gotoxy(struct pcp_device *pcpdev, uint8_t x, uint8_t y)
{
	int		ret;
	uint16_t	mask = BIT(7);

	if(x > 0x0F) {
		ERR("x=%d is too large\n", x);
		return -EINVAL;
	}
	if(y > 0x01) {
		ERR("y=%d is too large\n", y);
		return -EINVAL;
	}
	DBG("(%d, %d)\n", x, y);
	mask |= (y << 6);
	mask |= x;
	ret = pcp_lcd_op(pcpdev, mask);
	usleep(40);
	return ret;
}

int pcp_lcd_sendd(struct pcp_device *pcpdev, char ascii)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	DBG("ascii='%c'\n", ascii);
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_LCD_SENDD, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PCP, LCD_SENDD, ascii) = ascii;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	free_command(reply);
	return ret;
}

static int pcp_control(struct pcp_device *pcpdev, uint8_t bitmask, int on)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	uint8_t			control;
	uint8_t			prev;
	int			ret;

	DBG("bitmask=0x%02X on=%d\n", bitmask, on);
	assert(pcpdev != NULL);
	prev = control = pcpdev->control_byte;
	control = (on) ?  (control | bitmask) : (control & ~bitmask);
	if((cmd = pcp_new_command(pcpdev, PCP_CONTROL, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PCP, CONTROL, bitmask) = control;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0)
		goto out;
	pcpdev->control_byte = control;
out:
	free_command(reply);
	return (ret < 0) ? ret : prev;
}

int pcp_backlight(struct pcp_device *pcpdev, int on)
{
	int	ret;

	DBG("on=%d\n", on);
	ret = pcp_control(pcpdev, PCP_CONTROL_LIGHT, on);
	if(ret >= 0) {
		ret = ret & PCP_CONTROL_LIGHT;
		pcpdev->backlight_status = (uint8_t)on;
	}
	return (ret);
}

int pcp_getbacklightstatus(struct pcp_device *pcpdev)
{
	return pcpdev->backlight_status;
}

int pcp_buzzer(struct pcp_device *pcpdev, int on)
{
	int	ret;

	DBG("on=%d\n", on);
	ret = pcp_control(pcpdev, PCP_CONTROL_BUZZER, on);
	return (ret < 0) ? ret : ret & PCP_CONTROL_BUZZER;
}


int pcp_led(struct pcp_device *pcpdev, uint8_t bitmask)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	DBG("bitmask=0x%02X\n", bitmask);
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_LED_SEND, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PCP, LED_SEND, bitmask) = (bitmask << 1);
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	free_command(reply);
	return ret;
}

#define	MAX_KEYCODE_BIT	10

static int ffbit(uint16_t bitmask, int start)
{
	int	i;

	for(i = start; i < MAX_KEYCODE_BIT; i++) {
		if(bitmask & (1 << i))
			return i;
	}
	return -1;
}

const char *keycode_name(uint8_t keycode)
{
#define	KC_NAME(x)	[ KEYCODE_ ## x ] = "Key " #x
	const char	*names[] = {
		KC_NAME(0),
		KC_NAME(1),
		KC_NAME(2),
		KC_NAME(3),
		KC_NAME(4),
		KC_NAME(5),
		KC_NAME(6),
		KC_NAME(7),
		KC_NAME(8),
		KC_NAME(9),
		KC_NAME(DEL),
		KC_NAME(STAR),
		KC_NAME(ESC),
		KC_NAME(BAK),
		KC_NAME(MENU),
		KC_NAME(ENTER),
		KC_NAME(HOME),
		KC_NAME(REC),
		KC_NAME(TUNNEL),
		KC_NAME(DOWN),
		KC_NAME(IPADDR),
		KC_NAME(SOUTH),
		KC_NAME(NORTH),
		KC_NAME(WEST),
		KC_NAME(EAST),
	};
	if(keycode >= sizeof(names)/sizeof(names[0]))
		return NULL;
	return names[keycode];
}

static int pcp_key_xlate(const uint16_t keybuf[], uint8_t keycodes[], int bytenum)
{
	enum keycodes	key_table[5][5] = {
		{ KEYCODE_1,	KEYCODE_SOUTH,	KEYCODE_4,	KEYCODE_7,	KEYCODE_DEL	},
		{ KEYCODE_2,	KEYCODE_EAST,	KEYCODE_5,	KEYCODE_8,	KEYCODE_0,	},
		{ KEYCODE_3,	KEYCODE_ENTER,	KEYCODE_6,	KEYCODE_9,	KEYCODE_STAR,	},
		{ KEYCODE_ESC,	KEYCODE_WEST,	KEYCODE_BAK,	KEYCODE_REC,	KEYCODE_DOWN,	},
		{ KEYCODE_MENU,	KEYCODE_NORTH,	KEYCODE_HOME,	KEYCODE_TUNNEL,	KEYCODE_IPADDR,	},
	};
	int	i;

	for(i = 0; i < MAX_KEYBUF && i < bytenum / 2; i++) {
		uint16_t	bitmask = keybuf[i];
		int		row = ffbit(bitmask, 0);
		int		col = ffbit(bitmask, 5) - 5;
		uint8_t		kc;

		if(row >= 5 || col >= 5) {
			ERR("Bad key: pos=%2d bitmask=0x%X row=%d col=%d\n", i, bitmask, row, col);
			return -EINVAL;
		}
		if(row < 0 || col < 0) {
			ERR("No key: pos=%2d bitmask=0x%X row=%d col=%d\n", i, bitmask, row, col);
			return 0;
		}
		kc = key_table[row][col];
		DBG("%2d> [%d] %s\n", i, kc, keycode_name(kc));
		keycodes[i] = kc;
	}
	return bytenum / 2;
}

void pcp_show_keys(uint8_t keycodes[], int numkeys)
{
	int	i;

	if(numkeys <= 0)
		return;
	DBG("KEYS: %d\n", numkeys);
	for(i = 0; i < MAX_KEYBUF && i < numkeys; i++) {
		uint8_t		kc = keycodes[i];

		DBG("\t%2d> [%d] %s\n", i, kc, keycode_name(kc));
	}
}

int pcp_keyget(struct pcp_device *pcpdev, uint8_t keycodes[])
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		*keybuf;
	uint16_t		tx_seq;
	int			ret;

	DBG("\n");
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_KEY_GET, 0)) == NULL)
		return -ENOMEM;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0)
		goto out;
	if (ret < sizeof(reply->header)) {
		ERR("short answer (%d < %zd)\n", ret, sizeof(reply->header));
		goto out;
	}
	ret -= sizeof(reply->header);
	if (ret > MAX_KEYBUF * 2) {
		ERR("long answer (keybuf len %d > %d)\n",
			ret, MAX_KEYBUF * 2);
		goto out;
	}
	keybuf = CMD_FIELD(reply, PCP, KEY_GET_REPLY, keys);
	ret = pcp_key_xlate(keybuf, keycodes, ret);
	pcp_show_keys(keycodes, ret);
out:
	free_command(reply);
	return ret;
}

/**
 * Sets the fan speed.
 * Update the cached value in pcpdev->fan_speed[]
 * @returns the previous speed
 */
int pcp_fanspeed_set(struct pcp_device *pcpdev, uint8_t fan_number, uint8_t fan_speed)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	DBG("fan=%d speed=%d\n", fan_number, fan_speed);
	assert(pcpdev != NULL);
	if(fan_number >= FANS_NUM) {
		ERR("Illegal fan number %d (max %d)\n", fan_number, FANS_NUM);
		return -EINVAL;
	}
	if(fan_speed > FAN_MAX_SPEED) {
		fan_speed = FAN_MAX_SPEED;
		WARN("fan speed %d truncated to max %d\n", fan_speed, FAN_MAX_SPEED);
	}
	if((cmd = pcp_new_command(pcpdev, PCP_FAN_SET, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PCP, FAN_SET, fan_number) = fan_number;
	CMD_FIELD(cmd, PCP, FAN_SET, fan_speed) = fan_speed;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0)
		goto out;
	ret = pcpdev->fan_speed[fan_number];
	pcpdev->fan_speed[fan_number] = fan_speed;
	DBG("fan=%d speed=%d\n", fan_number, ret);
out:
	free_command(reply);
	return ret;
}

/**
 * Actually read the fan speed from the device.
 * Update the cached value in pcpdev->fan_speed[]
 * @returns the speed that was read
 */
int pcp_fanspeed_read(struct pcp_device *pcpdev, uint8_t fan_number)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	DBG("\n");
	assert(pcpdev != NULL);
	if(fan_number >= FANS_NUM) {
		ERR("Illegal fan number %d (max %d)\n", fan_number, FANS_NUM);
		return -EINVAL;
	}
	if((cmd = pcp_new_command(pcpdev, PCP_FAN_GET, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PCP, FAN_GET, fan_number) = fan_number;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0)
		goto out;
	pcpdev->fan_speed[fan_number] = CMD_FIELD(reply, PCP, FAN_GET_REPLY, fan_speed);
	ret = pcpdev->fan_speed[fan_number];
out:
	free_command(reply);
	return ret;
}

/**
 * @returns the fan speed
 * If the speed is not known, calls pcp_fanspeed_read()
 */
int pcp_fanspeed(struct pcp_device *pcpdev, uint8_t fan_number)
{
	int	speed;

	DBG("\n");
	assert(pcpdev != NULL);
	if(fan_number >= FANS_NUM) {
		ERR("Illegal fan number %d (max %d)\n", fan_number, FANS_NUM);
		return -EINVAL;
	}
	speed = pcpdev->fan_speed[fan_number];
	if(speed != FAN_SPEED_UNSET)
		return speed;
	return pcp_fanspeed_read(pcpdev, fan_number);
}

/**
 * Modifies the fan speed by a positive/negative delta.
 * Implementation calls pcp_fanspeed_set().
 * @returns the previous fan speed
 */
int pcp_fanspeed_change(struct pcp_device *pcpdev, uint8_t fan_number, int delta)
{
	int	speed;

	DBG("\n");
	assert(pcpdev != NULL);
	if(fan_number >= FANS_NUM) {
		ERR("Wrong fan number %d\n", fan_number);
		return -EINVAL;
	}
	if((speed = pcp_fanspeed(pcpdev, fan_number)) < 0) {
		return speed;
	}
	speed += delta;
	if(speed > FAN_MAX_SPEED)
		speed = FAN_MAX_SPEED;
	if(speed < 0)
		speed = 0;
	return pcp_fanspeed_set(pcpdev, fan_number, speed);
}

/**
 * Read the PSU status from the device.
 * @returns result of the command.
 */
int pcp_psu_read(struct pcp_device *pcpdev, int *psu_sts)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t				tx_seq;
	uint8_t					psu_stat;
	int						ret;

	DBG("\n");
	assert(pcpdev != NULL);
	
	*psu_sts = PCP_PSU_STATUS_DISCONNECTED;
	
	if((cmd = pcp_new_command(pcpdev, PCP_PSU_GET, 0)) == NULL)
		return -ENOMEM;
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	if(ret < 0)
		goto out;
	psu_stat = CMD_FIELD(reply, PCP, PSU_GET_REPLY, psu_status);
	*psu_sts = psu_stat & 0x03;
out:
	free_command(reply);
	return ret;
}

#if 0
static void zigbee_hexdump(const char *msg, const uint8_t *data, int len)
{
	int		i;

	fprintf(stderr, "%s (%d bytes):", msg, len);
	for(i = 0; i < len; i++)
		fprintf(stderr, " %02X", data[i]);
	fprintf(stderr, "\n");
}
#endif

int pcp_zigbee_send(struct pcp_device *pcpdev, uint8_t data[], uint8_t data_len)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		tx_seq;
	int			ret;

	DBG("\n");
	assert(pcpdev != NULL);
	if((cmd = pcp_new_command(pcpdev, PCP_XBEE_SEND, 0)) == NULL)
		return -ENOMEM;
	if(data_len > sizeof(CMD_FIELD(cmd, PCP, XBEE_SEND, data))) {
		ERR("Command too long\n");
		return -E2BIG;
	}
	//zigbee_hexdump(__FUNCTION__, data, data_len);
	memcpy(CMD_FIELD(cmd, PCP, XBEE_SEND, data), data, data_len);
	ret = pcp_process_command(pcpdev, cmd, &reply, &tx_seq);
	free_command(reply);
	return ret;
}

static uint8_t zigbee_chksum(const uint8_t *data, int len)
{
	int		i;
	unsigned int	chksum = 0;

	for(i = 0; i < len; i++) {
		chksum += data[i];
	}
	chksum &= 0xFF;
	return 0xFF - chksum;
}

int pcp_zigbee_lamp(struct pcp_device *pcpdev, uint8_t lamp_num, int on)
{
// off: 71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 31 05 6E
// on:  71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 31 04 6F

// off: 71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 30 05 6F
// on:  71 7E 00 10 17 01 00 00 00 00 00 00 00 00 FF FE 02 44 30 04 70
	struct zigbee_lamp {
		uint8_t		api_identifier;
		uint8_t		frame_id;
		uint8_t		dest_address[8];
		uint16_t	dest_network;
		uint8_t		cmd_options;
		uint8_t		cmd_name[2];
		uint8_t		cmd_param;
	} __attribute__((packed));
	struct zigbee_command {
		uint8_t		delimiter;
		uint16_t	len;
		union {
			uint8_t	data[sizeof(struct zigbee_lamp)];
			struct zigbee_lamp	lamp;
		} d;
		uint8_t		chksum;
	} __attribute__((packed));
	struct zigbee_command	cmd;

	DBG("pcp_zigbee_lamp(%d,%d)\n", lamp_num, on);
	cmd.delimiter = 0x7E;
	cmd.len = htons(sizeof(cmd.d.lamp));
	cmd.d.lamp.api_identifier = 0x17;
	cmd.d.lamp.frame_id = 0x01;
	memset(cmd.d.lamp.dest_address, 0, sizeof(cmd.d.lamp.dest_address));
	cmd.d.lamp.dest_network = htons(0xFFFE);
	cmd.d.lamp.cmd_options = 0x2;
	cmd.d.lamp.cmd_name[0] = 'D';
	cmd.d.lamp.cmd_name[1] = (lamp_num) ? '1' : '0';
	cmd.d.lamp.cmd_param = (on) ? 0x04 : 0x05;
	cmd.chksum = zigbee_chksum((uint8_t *)&cmd.d.lamp, sizeof(cmd.d.lamp));
	return pcp_zigbee_send(pcpdev, (uint8_t *)&cmd, sizeof(cmd));
}

/*
 * Wrappers
 */

struct pcp_device *pcp_new(struct xusb_iface *iface)
{
	struct pcp_device	*pcpdev;
	int			ret;

	DBG("\n");
	pcpdev = calloc(sizeof(*pcpdev), 1);
	if(!pcpdev) {
		ERR("Allocating PCP device memory failed\n");
		goto err;
	}
	memset(pcpdev->fan_speed, FAN_SPEED_UNSET, FANS_NUM);
	pcpdev->xtalk_base = xtalk_base_new_on_xusb(iface);
	pcpdev->xtalk_sync = xtalk_sync_new(pcpdev->xtalk_base);

	ret = xtalk_sync_set_protocol(pcpdev->xtalk_sync, &pcptalk_base);
	if(ret < 0) {
		ERR("PCP Protocol registration failed: %d\n", ret);
		goto err;
	}
	ret = xtalk_proto_query(pcpdev->xtalk_sync);
	if(ret < 0) {
		ERR("Protocol handshake failed: %d\n", ret);
		goto err;
	}
	pcpdev->panel_proto_version = ret;
	return pcpdev;

err:
	if (pcpdev)
		pcp_delete(pcpdev);
	return NULL;
}

void pcp_delete(struct pcp_device *pcpdev)
{
	DBG("\n");
	if(pcpdev) {
		xtalk_sync_delete(pcpdev->xtalk_sync);
		pcpdev->xtalk_base = NULL;
		free(pcpdev);
	}
}
