#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_modal.h>
#include <xpanel/xpw_background.h>
#include <xpanel/xpw_dialog.h>
#include <xpanel/xpw_editbox.h>
#include <xpanel/xpw_yesno.h>
#include <xpanel/xpw_ipaddr.h>
#include <xpanel/xpw_menu.h>
#include <xpanel/xpw_hotkeys.h>
#include <xpanel/xpw_displaybox.h>
#include <xpanel/xpw_passwd.h>

#define	DBG_MASK	0x10

/*------------------------------- Widget stack -------------------------------*/
#define	XPW_STACK_SIZE	10

static struct xpw	*xpw_stack[XPW_STACK_SIZE];
static int		xpw_sp = 0;

void dump_prefix(int depth, const char *prefix)
{
	int	i;

	for(i = 0; i < depth; i++)
		printf("    ");
	printf("%s", prefix);
}

void dump_xpw_stack(const char *msg)
{
	const struct xpw_class	*myclass;
	struct xpw		*widget;
	int			i;

	fprintf(stderr, "Dump widget stack (depth=%d): %s\n", xpw_sp, msg);
	for(i = 0; i < xpw_sp; i++) {
		widget = xpw_stack[i];
		myclass = xpw_getclass(widget->classid);
		fprintf(stderr, "\t----> frame %d\n", i);
		myclass->dump_widget(widget, 1, "");
	}
}

int xpw_stack_depth(void)
{
	return xpw_sp;
}

int xpw_stack_push(struct xpw *widget)
{
	if(xpw_sp >= XPW_STACK_SIZE)
		return -EOVERFLOW;
	xpw_stack[xpw_sp++] = widget;
	if(widget->populate)
		widget->populate(widget);
	widget_restart(widget);
	return 0;
}

int xpw_stack_pop(void)
{
	struct xpw *widget;

	if(xpw_sp <= 1)	/* Protect TOP */
		return -ENOENT;
	//dump_xpw_stack("pop start");
	xpw_sp--;
	widget = xpw_stack[xpw_sp];
	xpw_stack[xpw_sp] = NULL;
	//dump_xpw_stack("pop end");
	if(widget->cleanup)
		widget->cleanup(widget);
	return 0;
}

struct xpw *xpw_stack_peek(unsigned int depth)
{
	struct xpw	*widget;
	unsigned int	pos;

	if(depth > xpw_sp)	/* Protect TOP */
		return NULL;
	pos = xpw_sp - depth;
	widget = xpw_stack[pos];
	return widget;
}

struct xpw *xpw_stack_top(void)
{
	struct xpw	*widget;

	while(xpw_stack_pop() == 0)
		;
	widget = xpw_stack[0];
	if(widget->populate)
		widget->populate(widget);
	widget_restart(widget);
	return widget;
}

static const char *widgets_symbols[XPW_CLASS_LAST] = {
	[XPW_BACKGROUND]	= "^",
	[XPW_MENU]	= ">",
	[XPW_EDITBOX]	= "x",
};

int xpw_stack_string(char str[], int maxlen)
{
	int	i;

	memset(str, 0, maxlen);
	for(i = 1; i < xpw_sp && i < maxlen - 1; i++) {
		struct xpw	*widget;
		const char	*p;

		widget = xpw_stack[i];
		if(widget) {
			p = widgets_symbols[widget->classid];
			if(str) {
				str[i-1] = (p) ? *p : '?';
			}
		}
	}
	return xpw_sp;
}

struct xpw *xpw_current(void)
{
	struct xpw *widget;

	widget = xpw_stack[xpw_sp - 1];
	DBG("xpw_current = %p\n", widget);
	return widget;
}

static struct xpw	*current_toplevel = NULL;

struct xpw	*xpw_toplevel(void)
{
	return current_toplevel;
}

static int set_toplevel(struct xpw *toplevel)
{
	current_toplevel = toplevel;
	return 0;
}

/*------------------------------- Key processing -----------------------------*/
struct xpw *process_key(struct xpw *widget, enum keycodes key)
{
	struct xpw		*next_widget;

	if(key < 0 || key >= KEYCODE_UNUSED_LAST) {
		ERR("Bad key = 0x%X\n", key);
	}
	if((next_widget = process_hotkey(widget, key)) != NULL) {
		return next_widget;
	}
	if(widget == NULL) {
		ERR("NULL current. Going to top\n");
		widget = xpw_stack_top();
		
	}
	widget_call_handler(widget, key);
	next_widget = xpw_current();
	DBG("process_key(after): now in \"%s\"\n", next_widget->name);
	return next_widget;
}

/*------------------------------- Class handling -----------------------------*/
static const struct xpw_class	*widget_classes[XPW_CLASS_LAST];

#define	CLS_SETUP(c)	if((ret = xpw_ ## c ## _register()) < 0) return ret;
int xpw_init(struct xpw *toplevel)
{
	int	i;
	int	ret;

	CLS_SETUP(background);
	CLS_SETUP(modal);
	CLS_SETUP(dialog);
	CLS_SETUP(menu_notifier);
	CLS_SETUP(menu);
	CLS_SETUP(editbox);
	CLS_SETUP(ipaddr);
	CLS_SETUP(yesno);
	CLS_SETUP(displaybox);
	CLS_SETUP(passwd);
	for(i = XPW_UNUSED + 1; i < XPW_CLASS_LAST; i++) {
		const struct xpw_class	*c = widget_classes[i];

		if(c == NULL) {
			ERR("%s: Missing definitions of widget class #%d\n", __func__, i);
			return -EINVAL;
		}
		INFO("Registered class %s\n", c->name);
	}
	set_toplevel(toplevel);
	return 0;
}
#undef	CLS_SETUP

#define	CLS_SETUP(c)	xpw_ ## c ## _unregister()
int xpw_fini(void)
{
	int	i;

	CLS_SETUP(background);
	CLS_SETUP(modal);
	CLS_SETUP(dialog);
	CLS_SETUP(menu_notifier);
	CLS_SETUP(menu);
	CLS_SETUP(editbox);
	CLS_SETUP(ipaddr);
	CLS_SETUP(yesno);
	CLS_SETUP(displaybox);
	CLS_SETUP(passwd);
	for(i = XPW_UNUSED + 1; i < XPW_CLASS_LAST; i++) {
		const struct xpw_class	*c = widget_classes[i];

		if(c != NULL) {
			ERR("%s: widget class #%d is defined as \"%s\"\n", __func__, i, c->name);
			return -EINVAL;
		}
	}
	return 0;
}
#undef	CLS_SETUP

const struct xpw_class *xpw_getclass(enum xpw_classid classid)
{
	const struct xpw_class *xpw_class;

	if(classid >= XPW_CLASS_LAST) {
		ERR("Bad classid %d\n", classid);
		return NULL;
	}
	xpw_class = widget_classes[classid];
	if(!xpw_class) {
		ERR("Unknown class (classid = %d)\n", classid);
		return NULL;
	}
	//DBG("class = %s (%s)\n", xpw_class->name, xpw_current()->name);
	return xpw_class;
}

const struct xpw_class *xpw_myclass(struct xpw *xpw)
{
	return xpw_getclass(xpw->classid);
}

const char *xpw_myname(struct xpw *xpw)
{
	if(xpw)
		return xpw->name;
	return "(EMPTY)";
}

int xpw_class_register(const struct xpw_class *xpw_class)
{
	enum xpw_classid	classid = xpw_class->classid;

	if(classid < 0 || classid >= XPW_CLASS_LAST) {
		ERR("Bad classid %d\n", classid);
		return -EINVAL;
	}
	if(widget_classes[classid]) {
		ERR("Classid %d already registered (%s)\n", classid, xpw_class->name);
		return -EEXIST;
	}
	if(!xpw_class->dump_widget) {
		ERR("Classid %d (%s): No dump_widget() method defined\n", classid, xpw_class->name);
		return -EINVAL;
	}
	if(!xpw_class->redraw) {
		ERR("Classid %d (%s): No redraw() method defined\n", classid, xpw_class->name);
		return -EINVAL;
	}
	if(!xpw_class->restart) {
		ERR("Classid %d (%s): No restart() method defined\n", classid, xpw_class->name);
		return -EINVAL;
	}
	widget_classes[classid] = xpw_class;
	return 0;
}

int xpw_class_unregister(enum xpw_classid classid)
{
	widget_classes[classid] = NULL;
	return 0;
}

int xpw_set_callback(struct xpw *xpw, widget_notify_t notify, void *data)
{
	assert(xpw);
	xpw->notify = notify;
	xpw->user_data = data;
	return 0;
}

int xpw_set_dynamic(struct xpw *xpw, widget_func_t populate, widget_func_t cleanup)
{
	assert(xpw);
	xpw->populate = populate;
	xpw->cleanup = cleanup;
	return 0;
}

int xpw_set_modal(struct xpw *xpw, struct xpw_modal *modal, struct xpw *delegate)
{
	int			ret;

	if((ret = modal_construct(modal, xpw, delegate)) < 0) {
		ERR("%s: Failed (%d)\n", __func__, ret);
		return ret;
	}
	xpw->modal = modal;
	return 0;
}

void xpw_hotkey_set_mask(struct xpw *xpw, const uint64_t *new_mask, uint64_t *old_mask)
{
	if (old_mask)
		memcpy(old_mask, &xpw->hotkeys_masked, sizeof(xpw->hotkeys_masked));
	if (new_mask)
		memcpy(&xpw->hotkeys_masked, new_mask, sizeof(xpw->hotkeys_masked));
}

int xpw_hotkey_is_masked(struct xpw *xpw, enum keycodes key)
{
	return xpw->hotkeys_masked & (1 << key);
}

int xpw_construct(struct xpw *xpw, enum xpw_classid classid, const char *name)
{
	if(xpw_getclass(classid) == NULL) {
		ERR("%s: Missing class for classid=%d\n", __func__, classid);
		return -EINVAL;
	}
	xpw->classid = classid;
	strncpy(xpw->name, name,WIDGET_NAME_MAX_LEN);
	xpw->name[WIDGET_NAME_MAX_LEN]=0;
	xpw->notify = NULL;
	xpw->user_data = NULL;
	xpw->modal = NULL;
	xpw_set_dynamic(xpw, NULL, NULL);
	return 0;
}

/*
 * Convenience wrappers
 */
void widget_redraw(struct xpw_screen *screen, struct xpw *widget)
{
	const struct xpw_class	*myclass;

	if(!widget)
		return;
	myclass = xpw_getclass(widget->classid);
	assert(myclass);
	DBG("%s: [%s] \"%s\" (stack depth = %d)\n",
		__func__, myclass->name, widget->name, xpw_stack_depth());
	myclass->redraw(screen, widget);
}

void widget_restart(struct xpw *widget)
{
	const struct xpw_class	*myclass;

	if(widget) {
		myclass = xpw_getclass(widget->classid);
		assert(myclass);
		myclass->restart(widget);
	}
}

void handle_modal(struct xpw *widget)
{
	if(widget->modal) {
		modal_active(widget->modal, 1);
	}
}

void widget_notify(struct xpw *widget, enum keycodes key)
{
	if(!widget)
		return;
	if(widget->notify) {
		INFO("%s: \"%s\" (%s)\n", __func__, widget->name, keycode_name(key));
		widget->notify(widget, key, widget->user_data);
	}
}

void widget_call_handler(struct xpw *widget, enum keycodes key)
{
	const struct xpw_class	*myclass;
	widget_kbd_handler_t	handler;

	if(!widget)
		return;	/* ignore */
	myclass = xpw_getclass(widget->classid);
	assert(myclass);
	handler = myclass->kbd_handlers[key];
	if(!handler)
		return;	/* Ignore key */
	INFO("%s: [%s] \"%s\" (%s)\n", __func__, myclass->name, widget->name, keycode_name(key));
	handler(widget, key);
}
