#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_background.h>
#include <xpanel/xpw_hotkeys.h>

#define	DBG_MASK	0x80

static struct xpw_background	global_background;

static void dump_background_contents_line(struct xpw_background *background, int depth, const char *prefix, int line)
{
	char	*p;
	int	i;

	assert(line >= 0 && line < BACKGROUND_LINES);
	dump_prefix(depth, "");
	printf("%02d [", line);
	p = background->contents[line];
	for(i = 0; i < BACKGROUND_MAXWIDTH && *p; i++, p++) {
		putchar(*p);
	}
	printf("]\n");
}

static void dump_background(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_background	*background;

	assert(widget->classid == XPW_BACKGROUND);
	background = (struct xpw_background *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	depth++;
	dump_background_contents_line(background, depth, prefix, 0);
	dump_background_contents_line(background, depth, prefix, 1);
	if(background->toplevel) {
		const struct xpw_class	*menu_class;

		if((menu_class = xpw_myclass(WIDGET(background->toplevel))) == NULL) {
			ERR("Classless widget!. Aborting\n");
			exit(1);
		}
		menu_class->dump_widget(WIDGET(background->toplevel), depth, "");
	}
}

static void xpw_background_clear(struct xpw_background *xpw_background)
{
	int	i;

	for(i = 0; i < BACKGROUND_LINES; i++)
		memset(&xpw_background->contents[i], 0, sizeof(xpw_background->contents[i]));
}

static void redraw_background(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_background	*background;
	int		y;
	int		line;

	assert(widget->classid == XPW_BACKGROUND);
	background = (struct xpw_background *)widget;
	for(y = 0; y < 2; y++) {
		char	*p;

		line = (background->line_show + y) % BACKGROUND_LINES;
		p = background->contents[line];
		p += background->start_column;
		DBG("%d> (%d) [%s]\n", y, background->start_column, p);
		if(screen) {
			xpw_screen_gotoxy(screen, 0, y);
			xpw_screen_string(screen, p);
		}
	}
}

static void restart_background(struct xpw *widget)
{
	/* struct xpw_background	*background; */

	assert(widget->classid == XPW_BACKGROUND);
	/* background = (struct xpw_background *)widget; */
	DBG("\n");
}

static void background_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_background	*background = (struct xpw_background *)widget;
	struct xpw		*next_widget;
	int			line;

	//dump_xpw_stack("background_handler start");
	next_widget = NULL;
	switch(key) {
		case KEYCODE_ENTER:
		case KEYCODE_MENU:
			if(background->toplevel)
				next_widget = WIDGET(background->toplevel);
			if(next_widget)
				xpw_stack_push(next_widget);
			break;
		case KEYCODE_NORTH:
			line = background->line_show;
			if(--line <= 0)
				line = BACKGROUND_LINES - 1;
			background->line_show = line;
			break;
		case KEYCODE_SOUTH:
			line = background->line_show;
			if(++line >= BACKGROUND_LINES)
				line = 0;
			background->line_show = line;
			break;
		case KEYCODE_EAST:
			if(++background->start_column >= BACKGROUND_COLUMNS)
				background->start_column = BACKGROUND_COLUMNS;
			break;
		case KEYCODE_WEST:
			if(--background->start_column <= 0)
				background->start_column = 0;
			break;
		case KEYCODE_DEL:
			xpw_background_clear(background);
			break;
		default:
			break;
	}
	if(!next_widget)
		next_widget = widget;
	DBG("background_handler: \"%s\" -> \"%s\"\n", widget->name, next_widget->name);
}

static struct xpw_class	xpw_background_class = {
	.classid	= XPW_BACKGROUND,
	.name		= "Background",
	.kbd_handlers = {
		[KEYCODE_MENU]	background_handler,
		[KEYCODE_ENTER]	background_handler,
		[KEYCODE_NORTH]	background_handler,
		[KEYCODE_SOUTH]	background_handler,
		[KEYCODE_EAST]	background_handler,
		[KEYCODE_WEST]	background_handler,
		[KEYCODE_DEL]	background_handler,
	},
	.dump_widget	= dump_background,
	.redraw		= redraw_background,
	.restart	= restart_background,
};

struct xpw_background *xpw_background_get(void)
{
	return &global_background;
}

struct xpw *hotkey_background(enum keycodes key, struct xpw *widget)
{
	struct xpw		*top;
	struct xpw_background	*background;
	int			line;

	top = xpw_stack_top();
	assert(top);
	assert(top->classid == XPW_BACKGROUND);
	background = (struct xpw_background *)top;
	line = background->line_insert;
	line = (line + BACKGROUND_LINES - 2) % BACKGROUND_LINES;
	background->line_show = line;
	return top;
}

static int background_addone(struct xpw_background *background, const char *start, const char *end)
{
	int		y;
	int		len;
	char		tmpbuf[BACKGROUND_MAXWIDTH + 1];

	len = (end) ? end - start + 1 : strlen(start);
	if(len <= 0)
		return 0;			/* skip empty lines */
	if(len > BACKGROUND_MAXWIDTH)
		len = BACKGROUND_MAXWIDTH;	/* truncate */
	memcpy(tmpbuf, start, len);
	tmpbuf[len] = '\0';
	y = background->line_insert;
	INFO("%s: %02d -> \"%s\"\n", __func__, y, tmpbuf);
	memcpy(background->contents[y], start, len);
	background->contents[y][len] = '\0';
	background->line_show = (y == 0) ? 0 : y - 1;
	/* Next line */
	y = (y + 1) % BACKGROUND_LINES;
	background->line_insert = y;
	background->contents[y][0] = '\0';
	return 0;
}

int xpw_background_addstring(struct xpw_background *background, const char *str)
{
	const char	*curr;
	const char	*prev;

	INFO("%s: \"%s\"\n", __func__, str);
	prev = str;
	for(curr = strchr(prev, '\n'); curr; curr = strchr(prev, '\n')) {
		background_addone(background, prev, curr - 1);
		prev = curr + 1;
	}
	background_addone(background, prev, NULL);
	return 0;
}

static void fill_logo(struct xpw_background *xpw_background, const char *logo)
{
	if(logo == NULL)
		xpw_background_addstring(xpw_background, "Xorcom Panel");
	else
		xpw_background_addstring(xpw_background, logo);
}

int xpw_background_construct(struct xpw_background *xpw_background, struct xpw *toplevel, const char *logo)
{
	xpw_construct(&xpw_background->myself, XPW_BACKGROUND, "background");
	if(xpw_background->toplevel) {
		ERR("Toplevel widget is already set\n");
		return -EBUSY;
	}
	xpw_background->line_insert = 0;
	xpw_background->line_show = 0;
	xpw_background->start_column = 0;
	xpw_background->toplevel = toplevel;
	xpw_background_clear(xpw_background);
	fill_logo(xpw_background, logo);
	return 0;
}

int xpw_background_startup(struct xpw *toplevel, const char *logo)
{
	int	ret;

	assert(toplevel);
	xpw_background_construct(&global_background, toplevel, logo);
	if((ret = xpw_hotkey_set(KEYCODE_HOME, NULL, hotkey_background)) < 0)
		return ret;
	return 0;
}

int xpw_background_register(void)
{
	return xpw_class_register(&xpw_background_class);
}

void xpw_background_unregister(void)
{
	xpw_hotkey_unset(KEYCODE_HOME);
	xpw_class_unregister(XPW_BACKGROUND);
	global_background.toplevel = NULL;
}
