#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_dialog.h>

#define	DBG_MASK	0x80

#define	PROMPT_POS	(-1)

static void dump_dialog(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_dialog	*dialog;
	int		i;

	assert(widget->classid == XPW_DIALOG);
	dialog = (struct xpw_dialog *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\": \"%s\"\n", widget->name, dialog->prompt);
	depth++;
	for(i = 0; i < MAX_DIALOG_ITEMS; i++) {
		const struct xpw_class	*the_class;
		struct xpw		*widget;
		char			prefixbuf[BUFSIZ];

		widget = dialog->entries[i];
		if(!widget)
			continue;
		the_class = xpw_getclass(widget->classid);
		snprintf(prefixbuf, sizeof(prefixbuf),
				"%c%d> ", (i == dialog->curr_entry) ? '+' : ' ', i);
		dump_prefix(depth, prefixbuf);
		printf("[%s] %s:\n", the_class->name, widget->name);
		the_class->dump_widget(widget, depth, "    ");
	}
}

static void redraw_dialog(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	int			pos;

	assert(widget->classid == XPW_DIALOG);
	dialog = (struct xpw_dialog *)widget;
	pos = dialog->curr_entry;
	if(pos >= 0 && pos <= MAX_DIALOG_ITEMS) {
		struct xpw	*subwidget = dialog->entries[pos];

		INFO("%s: pos=%d\n", __func__, pos);
		assert(subwidget);
		xpw_screen_nesting(screen, 1);
		widget_redraw(screen, subwidget);
		xpw_screen_nesting(screen, 0);
	} else {
		INFO("%s: Title = %s\n", __func__, dialog->prompt);
		xpw_screen_gotoxy(screen, 0, 1);	/* bottom line */
		xpw_screen_string(screen, "");
		xpw_screen_gotoxy(screen, 0, 0);
		xpw_screen_string(screen, dialog->prompt);
	}
}

static void restart_dialog(struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	int			i;

	DBG("%s\n", __func__);
	assert(widget->classid == XPW_DIALOG);
	dialog = (struct xpw_dialog *)widget;
	dialog->curr_entry = PROMPT_POS;
	for(i = 0; i < MAX_DIALOG_ITEMS; i++) {
		if(dialog->entries[i])
			widget_restart(dialog->entries[i]);
	}
}

static void dialog_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_dialog	*dialog = (struct xpw_dialog *)widget;
	struct xpw		*entry;
	struct xpw		*next_entry = NULL;
	int			pos;

	// dump_xpw_stack("dialog_handler start");
	// dump_dialog(widget, 0, ">>>>");
	pos = dialog->curr_entry;
	if(pos == PROMPT_POS || pos < 0 || pos >= MAX_DIALOG_ITEMS)
		entry = NULL;
	else
		entry = dialog->entries[pos];
	switch(key) {
		case KEYCODE_NORTH:
			pos = (pos <= 0) ? PROMPT_POS : pos - 1;
			break;		/* Finished this key */
		case KEYCODE_SOUTH:
			if(pos == PROMPT_POS && entry) {
				pos = 0;
			} else if(pos < MAX_DIALOG_ITEMS - 1) {
				if(dialog->entries[pos + 1])
					pos++;
			}
			break;		/* Finished this key */
		case KEYCODE_ENTER:
			INFO("%s: ENTER '%s'\n", __func__, xpw_myname(WIDGET(dialog)));
			handle_modal(WIDGET(dialog));
			widget_notify(entry, key);
			break;
		case KEYCODE_ESC:
			INFO("%s: ESC '%s'\n", __func__, xpw_myname(WIDGET(dialog)));
			xpw_stack_pop();
			//dump_xpw_stack("dialog exit");
			break;
		default:
			widget_call_handler(entry, key);
	}
	dialog->curr_entry = pos;
	if(pos != PROMPT_POS) {
		next_entry = dialog->entries[pos];
	}
	INFO("%s(%d): \"%s\" -> \"%s\"\n", __func__, pos,
		xpw_myname(entry), xpw_myname(next_entry));
}

static struct xpw_class	xpw_dialog_class = {
	.classid	= XPW_DIALOG,
	.name		= "Dialog",
	.kbd_handlers = {
		[KEYCODE_0]	dialog_handler,
		[KEYCODE_1]	dialog_handler,
		[KEYCODE_2]	dialog_handler,
		[KEYCODE_3]	dialog_handler,
		[KEYCODE_4]	dialog_handler,
		[KEYCODE_5]	dialog_handler,
		[KEYCODE_6]	dialog_handler,
		[KEYCODE_7]	dialog_handler,
		[KEYCODE_8]	dialog_handler,
		[KEYCODE_9]	dialog_handler,
		[KEYCODE_STAR]	dialog_handler,
		[KEYCODE_DEL]	dialog_handler,
		[KEYCODE_WEST]	dialog_handler,
		[KEYCODE_EAST]	dialog_handler,
		[KEYCODE_SOUTH]	dialog_handler,
		[KEYCODE_NORTH]	dialog_handler,
		[KEYCODE_ESC]	dialog_handler,
		[KEYCODE_ENTER]	dialog_handler,
	},
	.dump_widget	= dump_dialog,
	.redraw		= redraw_dialog,
	.restart	= restart_dialog,
};

int xpw_dialog_addwidget(struct xpw_dialog *xpw_dialog, struct xpw *widget)
{
	int		i;

	for(i = 0; i < MAX_DIALOG_ITEMS; i++) {
		if(xpw_dialog->entries[i] == NULL)
			break;
	}
	if(i == MAX_DIALOG_ITEMS)
		return -ENOSPC;
	DBG("%s[%d]: %s\n", __func__, i, widget->name);
	xpw_dialog->entries[i] = widget;
	return 0;
}

int xpw_dialog_delwidget(struct xpw_dialog *xpw_dialog, struct xpw *widget)
{
	int		i;

	for(i = 0; i < MAX_DIALOG_ITEMS; i++) {
		if(xpw_dialog->entries[i] == widget) {
			xpw_dialog->entries[i] = NULL;
			return 0;
		}
	}
	return -ENOENT;
}

int xpw_dialog_construct(struct xpw_dialog *xpw_dialog, const char prompt[])
{
	xpw_construct(&xpw_dialog->myself, XPW_DIALOG, prompt);
	snprintf(xpw_dialog->prompt, sizeof(xpw_dialog->prompt), "%s", prompt);
	memset(xpw_dialog->entries, 0, sizeof(xpw_dialog->entries));
	xpw_dialog->curr_entry = PROMPT_POS;
	return 0;
}

int xpw_dialog_register(void)
{
	return xpw_class_register(&xpw_dialog_class);
}

void xpw_dialog_unregister(void)
{
	xpw_class_unregister(XPW_DIALOG);
}

