#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_editbox.h>

#define	DBG_MASK	0x80

static void dump_editbox(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_editbox	*editbox;
	char			*p;
	int			i;
	int			inside = 0;

	assert(widget->classid == XPW_EDITBOX);
	editbox = (struct xpw_editbox *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	depth++;
	dump_prefix(depth, "[");
	p = editbox->contents;
	for(i = 0; i < EDITBOX_MAXWIDTH && *p; i++, p++) {
		if(i == editbox->cursor_pos) {
			inside = 1;
			putchar('|');
		}
		putchar(*p);
		if(i == editbox->cursor_pos)
			putchar('|');
	}
	if(!inside)
		printf("|.|");
	printf("]\n");
}

int xpw_editbox_construct(struct xpw_editbox	*xpw_ebox, const char *prompt)
{
	xpw_construct(&xpw_ebox->myself, XPW_EDITBOX, prompt);
	snprintf(xpw_ebox->prompt, sizeof(xpw_ebox->prompt), "%s:", prompt);
	memset(xpw_ebox->contents, 0, sizeof(xpw_ebox->contents));
	xpw_ebox->cursor_pos = 0;
	return 0;	
}

static void redraw_editbox(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_editbox	*editbox;

	assert(widget->classid == XPW_EDITBOX);
	editbox = (struct xpw_editbox *)widget;
	DBG("[%s]\n", editbox->contents);
	xpw_screen_gotoxy(screen, 0, 0);
	xpw_screen_string(screen, editbox->prompt);
	xpw_screen_gotoxy(screen, 0, 1);
	xpw_screen_string(screen, editbox->contents);
	xpw_screen_gotoxy(screen, editbox->cursor_pos, 1);
}

static void restart_editbox(struct xpw *widget)
{
	struct xpw_editbox	*editbox;

	assert(widget->classid == XPW_EDITBOX);
	editbox = (struct xpw_editbox *)widget;
	DBG("[%s]\n", editbox->contents);
	memset(editbox->contents, 0, sizeof(editbox->contents));
	editbox->cursor_pos = 0;
	return;
}

static void append_char(struct xpw_editbox *editbox, char ch)
{
	int	pos = editbox->cursor_pos;

	if(pos < 0 || pos >= EDITBOX_MAXWIDTH) {
		ERR("%s: illegal cursor position %d\n", __FUNCTION__, pos);
		return;
	}
	for(pos = 0; pos < EDITBOX_MAXWIDTH && editbox->contents[pos] != '\0'; pos++)
		;
	editbox->contents[pos++] = ch;
	editbox->contents[pos] = '\0';
	editbox->cursor_pos = pos;
	DBG("%s: [%d] '%s'\n", __FUNCTION__, pos, editbox->contents);
}

static void editbox_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_editbox	*editbox = (struct xpw_editbox *)widget;
	struct xpw		*next_widget;

	//dump_xpw_stack("editbox_handler start");
	next_widget = NULL;
	switch(key) {
		case KEYCODE_0:
		case KEYCODE_1:
		case KEYCODE_2:
		case KEYCODE_3:
		case KEYCODE_4:
		case KEYCODE_5:
		case KEYCODE_6:
		case KEYCODE_7:
		case KEYCODE_8:
		case KEYCODE_9:
		case KEYCODE_STAR:
			if(editbox->cursor_pos < EDITBOX_MAXWIDTH) {
				char	ch = 0;

				if(key >= KEYCODE_0 && key <= KEYCODE_9)
					ch = '0' + (key - KEYCODE_0);
				else if(key == KEYCODE_STAR)
					ch = '*';
				if(editbox->contents[editbox->cursor_pos] == '\0') {
					append_char(editbox, ch);
				} else {
					editbox->contents[editbox->cursor_pos++] = ch;
				}
			}
			break;
		case KEYCODE_DEL:
			{
				char	*p = &editbox->contents[editbox->cursor_pos];
				int	rest = strlen(p);

				DBG("DEL(%d)\n", rest);
				memmove(p, p+1, rest);
			}
			break;
		case KEYCODE_WEST:
			if(editbox->cursor_pos > 0)
				editbox->cursor_pos--;
			break;
		case KEYCODE_EAST:
			if(editbox->cursor_pos < EDITBOX_MAXWIDTH) {
				if(editbox->contents[editbox->cursor_pos] == '\0') {
					append_char(editbox, ' ');
				} else {
					editbox->cursor_pos++;
				}
			}
			break;
		case KEYCODE_ENTER:
			editbox->contents[editbox->cursor_pos] = '\0';
			editbox->cursor_pos = 0;
			INFO("EDITBOX(%s): Got '%s'\n", xpw_myname(WIDGET(editbox)), editbox->contents);
			widget_notify(widget, KEYCODE_ENTER);
			next_widget = xpw_current();
		case KEYCODE_ESC:
			if(xpw_stack_pop() == -ENOENT) {
				WARN("Hitting top. Staying there.\n");
			}
			next_widget = xpw_current();
			break;
		default:
			break;
	}
	if(!next_widget)
		next_widget = widget;
	DBG("editbox_handler: \"%s\" -> \"%s\"\n", widget->name, next_widget->name);
}

static struct xpw_class	xpw_editbox_class = {
	.classid	= XPW_EDITBOX,
	.name		= "Editbox",
	.kbd_handlers = {
		[KEYCODE_0]	editbox_handler,
		[KEYCODE_1]	editbox_handler,
		[KEYCODE_2]	editbox_handler,
		[KEYCODE_3]	editbox_handler,
		[KEYCODE_4]	editbox_handler,
		[KEYCODE_5]	editbox_handler,
		[KEYCODE_6]	editbox_handler,
		[KEYCODE_7]	editbox_handler,
		[KEYCODE_8]	editbox_handler,
		[KEYCODE_9]	editbox_handler,
		[KEYCODE_STAR]	editbox_handler,
		[KEYCODE_DEL]	editbox_handler,
		[KEYCODE_WEST]	editbox_handler,
		[KEYCODE_EAST]	editbox_handler,
		[KEYCODE_ESC]	editbox_handler,
		[KEYCODE_ENTER]	editbox_handler,
	},
	.dump_widget	= dump_editbox,
	.redraw		= redraw_editbox,
	.restart	= restart_editbox,
};

int xpw_editbox_register(void)
{
	return xpw_class_register(&xpw_editbox_class);
}

void xpw_editbox_unregister(void)
{
	xpw_class_unregister(XPW_EDITBOX);
}
