#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_ipaddr.h>
#include <xpanel/xpw_hotkeys.h>

#define	DBG_MASK	0x80

static void dump_ipaddr(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_ipaddr	*ip;
	char		*p;
	int		i;
	int		inside = 0;

	assert(widget->classid == XPW_IPADDR);
	ip = (struct xpw_ipaddr *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	depth++;
	dump_prefix(depth, "[");
	p = ip->contents;
	for(i = 0; i < XPW_IPADDR_MAXWIDTH && *p; i++, p++) {
		if(i == ip->cursor_pos) {
			inside = 1;
			putchar('|');
		}
		putchar(*p);
		if(i == ip->cursor_pos)
			putchar('|');
	}
	if(!inside)
		printf("|.|");
	printf("]\n");
}

static void redraw_ipaddr(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_ipaddr	*ip;

	assert(widget->classid == XPW_IPADDR);
	ip = (struct xpw_ipaddr *)widget;
	DBG("[%s]\n", ip->contents);
	xpw_screen_gotoxy(screen, 0, 0);
	xpw_screen_string(screen, ip->prompt);
	xpw_screen_gotoxy(screen, 0, 1);
	xpw_screen_string(screen, ip->contents);
	xpw_screen_gotoxy(screen, ip->cursor_pos, 1);
}

static void restart_ipaddr(struct xpw *widget)
{
	struct xpw_ipaddr	*ip;
	const char	*ip_str;

	assert(widget->classid == XPW_IPADDR);
	ip = (struct xpw_ipaddr *)widget;
	ip_str = inet_ntop(AF_INET, &ip->inet_addr, ip->contents, sizeof(ip->contents));
	if(ip_str) {
		DBG("setting to '%s' (%zd chars)\n", ip_str, strlen(ip_str));
	} else {
		ERR("inet_ntop() failed.\n");
		memset(ip->contents, 0, sizeof(ip->contents));
	}
	ip->cursor_pos = 0;
}

static void append_char(struct xpw_ipaddr *ip, char ch)
{
	int	pos = ip->cursor_pos;

	if(pos < 0 || pos >= XPW_IPADDR_MAXWIDTH) {
		ERR("%s: illegal cursor position %d\n", __FUNCTION__, pos);
		return;
	}
	for(pos = 0; pos < XPW_IPADDR_MAXWIDTH && ip->contents[pos] != '\0'; pos++)
		;
	ip->contents[pos++] = ch;
	ip->contents[pos] = '\0';
	ip->cursor_pos = pos;
	DBG("%s: [%d] '%s'\n", __FUNCTION__, pos, ip->contents);
}

static void ipaddr_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_ipaddr	*ip = (struct xpw_ipaddr *)widget;
	struct xpw		*next_widget;

	//dump_xpw_stack("ipaddr_handler start");
	next_widget = NULL;
	switch(key) {
		case KEYCODE_0:
		case KEYCODE_1:
		case KEYCODE_2:
		case KEYCODE_3:
		case KEYCODE_4:
		case KEYCODE_5:
		case KEYCODE_6:
		case KEYCODE_7:
		case KEYCODE_8:
		case KEYCODE_9:
		case KEYCODE_STAR:
			if(ip->cursor_pos < XPW_IPADDR_MAXWIDTH) {
				char	ch = 0;

				if(key >= KEYCODE_0 && key <= KEYCODE_9)
					ch = '0' + (key - KEYCODE_0);
				else if(key == KEYCODE_STAR)
					ch = '.';
				if(ip->contents[ip->cursor_pos] == '\0') {
					append_char(ip, ch);
				} else {
					ip->contents[ip->cursor_pos++] = ch;
				}
			}
			break;
		case KEYCODE_DEL:
			{
				char	*p = &ip->contents[ip->cursor_pos];
				int	rest = strlen(p);

				DBG("DEL(%d)\n", rest);
				memmove(p, p+1, rest);
			}
			break;
		case KEYCODE_WEST:
			if(ip->cursor_pos > 0)
				ip->cursor_pos--;
			break;
		case KEYCODE_EAST:
			if(ip->cursor_pos < XPW_IPADDR_MAXWIDTH) {
				if(ip->contents[ip->cursor_pos] != '\0') {
					ip->cursor_pos++;
				} /* else do nothing */
			}
			break;
		case KEYCODE_ENTER:
			INFO("%s: Read '%s'\n", xpw_myname(WIDGET(ip)), ip->contents);
			next_widget = xpw_current();
			break;
		case KEYCODE_ESC:
			if(xpw_stack_pop() == -ENOENT) {
				WARN("Hitting top. Staying there.\n");
			}
			next_widget = xpw_current();
			break;
		default:
			break;
	}
	if(!next_widget)
		next_widget = widget;
	DBG("ipaddr_handler: \"%s\" -> \"%s\"\n", widget->name, next_widget->name);
}

static struct xpw_class	xpw_ipaddr_class = {
	.classid	= XPW_IPADDR,
	.name		= "IPAddr",
	.kbd_handlers = {
		[KEYCODE_0]	ipaddr_handler,
		[KEYCODE_1]	ipaddr_handler,
		[KEYCODE_2]	ipaddr_handler,
		[KEYCODE_3]	ipaddr_handler,
		[KEYCODE_4]	ipaddr_handler,
		[KEYCODE_5]	ipaddr_handler,
		[KEYCODE_6]	ipaddr_handler,
		[KEYCODE_7]	ipaddr_handler,
		[KEYCODE_8]	ipaddr_handler,
		[KEYCODE_9]	ipaddr_handler,
		[KEYCODE_STAR]	ipaddr_handler,
		[KEYCODE_DEL]	ipaddr_handler,
		[KEYCODE_WEST]	ipaddr_handler,
		[KEYCODE_EAST]	ipaddr_handler,
		[KEYCODE_ESC]	ipaddr_handler,
		[KEYCODE_ENTER]	ipaddr_handler,
	},
	.dump_widget	= dump_ipaddr,
	.redraw		= redraw_ipaddr,
	.restart	= restart_ipaddr,
};

int xpw_ipaddr_set(struct xpw *widget, uint32_t addr)
{
	struct xpw_ipaddr	*ip;

	assert(widget->classid == XPW_IPADDR);
	ip = (struct xpw_ipaddr *)widget;
	ip->inet_addr.s_addr = addr;
	return 0;
}

#if 0
static struct xpw *hotkey_ipaddr(enum keycodes key, struct xpw *widget)
{
	struct xpw_ipaddr	*xpw_ipaddr;

	assert(widget->classid == XPW_IPADDR);
	xpw_ipaddr = (struct xpw_ipaddr *)widget;
	return widget;
}
#endif

void xpw_ipaddr_result(struct xpw_ipaddr *xpw_ipaddr, char output[], size_t outlen)
{
	assert(xpw_ipaddr);
	snprintf(output, outlen, "%s", xpw_ipaddr->contents);
}

int xpw_ipaddr_construct(struct xpw_ipaddr *xpw_ipaddr, const char *prompt, uint32_t net_addr)
{
	xpw_construct(&xpw_ipaddr->myself, XPW_IPADDR, prompt);
	xpw_ipaddr->inet_addr.s_addr = net_addr;
	snprintf(xpw_ipaddr->prompt, sizeof(xpw_ipaddr->prompt), "%s:", prompt);
	memset(xpw_ipaddr->contents, 0, sizeof(xpw_ipaddr->contents));
	xpw_ipaddr->cursor_pos = 0;
	return 0;
}

int xpw_ipaddr_register(void)
{
	return xpw_class_register(&xpw_ipaddr_class);
}

void xpw_ipaddr_unregister(void)
{
	xpw_hotkey_unset(KEYCODE_IPADDR);
	xpw_class_unregister(XPW_IPADDR);
}
