#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_menu.h>
#include <xpanel/xpw_hotkeys.h>

#define	DBG_MASK	0x80

static void dump_menu(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_menu		*menu;
	char			prefixbuf[BUFSIZ];
	int			i;

	assert(widget->classid == XPW_MENU);
	menu = (struct xpw_menu *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	depth++;
	for(i = 0; i < MAX_MENU_ITEMS; i++) {
		struct xpw_menu_entry	*entry;

		entry = menu->entries[i];
		if(entry && entry->text) {
			const struct xpw_class	*the_class;

			widget = entry->widget;
			snprintf(prefixbuf, sizeof(prefixbuf),
					"%c%d> ", (i == menu->curr_entry) ? '+' : ' ', i+1);
			dump_prefix(depth, prefixbuf);
			printf("[%s]:\n", entry->text);
			if(widget) {
				the_class = xpw_getclass(widget->classid);
				the_class->dump_widget(widget, depth, "    ");
			} else {
				dump_prefix(depth, "     <NULL>");
			}
		}
	}
}

static void draw_menu_entry(struct xpw_screen *screen, struct xpw_menu *menu, int y, int curr)
{
	struct xpw_menu_entry	*entry;
	char			buf[BUFSIZ];

	entry = menu->entries[curr + y];
	if(entry)
		snprintf(buf, sizeof(buf), "%d. %s", curr + y + 1, entry->text);
	else
		buf[0] = '\0';
	INFO("%s\n", buf);
	if(screen) {
		xpw_screen_gotoxy(screen, 0, y);
		xpw_screen_string(screen, buf);
	}
}

static void redraw_menu(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_menu		*menu;
	int			curr;
	int			y;

	assert(widget->classid == XPW_MENU);
	menu = (struct xpw_menu *)widget;
	curr = menu->curr_entry;
	for(y = 0; y <= 1; y++) {
		draw_menu_entry(screen, menu, y, curr);
	}
	if(screen)
		xpw_screen_gotoxy(screen, 0, 0);
}

static void restart_menu(struct xpw *widget)
{
	struct xpw_menu		*menu;

	assert(widget->classid == XPW_MENU);
	menu = (struct xpw_menu *)widget;
	menu->curr_entry = 0;
}

static void menu_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_menu	*menu = (struct xpw_menu *)widget;
	struct xpw_menu_entry	*next_entry;
	struct xpw		*next_widget;

	//dump_xpw_stack("menu_handler start");
	next_widget = NULL;
	switch(key) {
		case KEYCODE_1:
		case KEYCODE_2:
		case KEYCODE_3:
		case KEYCODE_4:
		case KEYCODE_5:
		case KEYCODE_6:
		case KEYCODE_7:
		case KEYCODE_8:
		case KEYCODE_9:
			/* Direct jump */
			{
				int	i;

				i = key - KEYCODE_1;
				if(menu->entries[i])
					menu->curr_entry = i;
			}
			break;
		case KEYCODE_SOUTH:
			{
				int	nxt = menu->curr_entry + 1;

				if(nxt < MAX_MENU_ITEMS && menu->entries[nxt])
					menu->curr_entry = nxt;
			}
			break;
		case KEYCODE_NORTH:
			if(menu->curr_entry > 0)
				menu->curr_entry--;
			break;
		case KEYCODE_ESC:
			if(xpw_stack_pop() == -ENOENT) {
				WARN("Hitting top. Staying there.\n");
			}
			next_widget = xpw_current();
			break;
		case KEYCODE_ENTER:
			next_entry = menu->entries[menu->curr_entry];
			if(next_entry)
				next_widget = WIDGET(next_entry->widget);
			if(next_widget)
				xpw_stack_push(next_widget);
			break;
		default:
			break;
	}
	if(!next_widget)
		next_widget = widget;
	DBG("menu_handler: \"%s\" -> \"%s\"\n", widget->name, next_widget->name);
}

static void menu_notifier_handler(struct xpw *widget, enum keycodes key)
{
	//dump_xpw_stack("menu_notifier_handler start");
	switch(key) {
		case KEYCODE_ENTER:
			INFO("%s: done\n", __func__);
			xpw_stack_pop();
			break;
		default:
			break;
	}
	DBG("menu_notifier_handler: \"%s\" done\n", widget->name);
}

static void dump_menu_notifier(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_menu_notifier	*mn;

	assert(widget->classid == XPW_MENU_NOTIFIER);
	mn = (struct xpw_menu_notifier *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	printf("\"%s\":\n", mn->logo);
	depth++;
}

static void redraw_menu_notifier(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_menu_notifier	*mn;

	INFO("%s\n", __func__);
	assert(widget->classid == XPW_MENU_NOTIFIER);
	mn = (struct xpw_menu_notifier *)widget;
	if(screen) {
		xpw_screen_gotoxy(screen, 0, 0);
		if(mn->logo[0] == 0x00)
			xpw_screen_string(screen, widget->name);
		else
			xpw_screen_string(screen, mn->logo);
		xpw_screen_gotoxy(screen, 0, 1);
		xpw_screen_string(screen, "Press <ENTER>");
	}
}

static void restart_menu_notifier(struct xpw *widget)
{
	/* struct xpw_menu_notifier	*mn; */

	INFO("%s\n", __func__);
	assert(widget->classid == XPW_MENU_NOTIFIER);
	/* mn = (struct xpw_menu_notifier *)widget; */
	widget_notify(widget, KEYCODE_ENTER);
}

static struct xpw_class xpw_menu_notifier_class = {
	.classid	= XPW_MENU_NOTIFIER,
	.name		= "MenuNotifier",
	.kbd_handlers = {
		[KEYCODE_ENTER]	menu_notifier_handler,
	},
	.dump_widget	= dump_menu_notifier,
	.redraw		= redraw_menu_notifier,
	.restart	= restart_menu_notifier,
};

static struct xpw_class	xpw_menu_class = {
	.classid	= XPW_MENU,
	.name		= "Menu",
	.kbd_handlers = {
		[KEYCODE_1]	menu_handler,
		[KEYCODE_2]	menu_handler,
		[KEYCODE_3]	menu_handler,
		[KEYCODE_4]	menu_handler,
		[KEYCODE_5]	menu_handler,
		[KEYCODE_6]	menu_handler,
		[KEYCODE_7]	menu_handler,
		[KEYCODE_8]	menu_handler,
		[KEYCODE_9]	menu_handler,
		[KEYCODE_SOUTH]	menu_handler,
		[KEYCODE_NORTH]	menu_handler,
		[KEYCODE_ESC]	menu_handler,
		[KEYCODE_ENTER]	menu_handler,
	},
	.dump_widget	= dump_menu,
	.redraw		= redraw_menu,
	.restart	= restart_menu,
};

static struct xpw *hotkey_toplevel(enum keycodes key, struct xpw *widget)
{
	struct xpw	*top;

	DBG("(from %s)\n", (widget) ? widget->name : "nowhere");
	top = xpw_stack_top();
	assert(top);
	top = xpw_toplevel();
	assert(top);
	xpw_stack_push(top);
	return top;
}

int xpw_menu_entry_construct(struct xpw_menu_entry *me, struct xpw *widget, char text[])
{
	assert(widget != NULL);
	assert(text != NULL);
	me->widget = widget;
	me->text = text;
	return 0;
}

void xpw_menu_entry_clear(struct xpw_menu_entry *me)
{
	me->widget = NULL;
	me->text = NULL;
}

int xpw_menu_notifier_construct(struct xpw_menu_notifier *notifier, const char prompt[], const char logo[])
{
	struct xpw	*xpw;

	xpw = WIDGET(notifier);
	xpw_construct(xpw, XPW_MENU_NOTIFIER, prompt);
	xpw->notify = NULL;
	xpw->user_data = NULL;
	xpw_menu_notifier_set_logo(notifier, logo);
	return 0;
}

void xpw_menu_notifier_set_logo(struct xpw_menu_notifier *notifier, const char logo[])
{
	if(NULL != logo) {
		strncpy(notifier->logo, logo, MAX_MENU_NOTIFIER_LOGO_LEN);
		notifier->logo[MAX_MENU_NOTIFIER_LOGO_LEN] = 0;
	}
	else
		notifier->logo[0] = 0; 
}

int xpw_menu_addentry(struct xpw_menu *xpw_menu, struct xpw_menu_entry *menu_entry)
{
	int		i;

	for(i = 0; i < MAX_MENU_ITEMS; i++) {
		if(xpw_menu->entries[i] == NULL)
			break;
	}
	if(i == MAX_MENU_ITEMS)
		return -ENOSPC;
	assert(menu_entry->widget != NULL);
	assert(menu_entry->text != NULL);
	xpw_menu->entries[i] = menu_entry;
	return 0;
}

int xpw_menu_delentry(struct xpw_menu *xpw_menu, int pos)
{
	int			i;

	assert(xpw_menu);
	assert(pos >= 0);
	assert(pos < MAX_MENU_ITEMS);
	for(i = pos; i < MAX_MENU_ITEMS - 1; i++) {
		xpw_menu->entries[i] = xpw_menu->entries[i+1];
	}
	xpw_menu->entries[i] = NULL;
	while(pos > 0 && xpw_menu->entries[pos] == NULL)
		pos--;
	if(xpw_menu->curr_entry >= pos)
		xpw_menu->curr_entry = pos;
	return 0;
}

int xpw_menu_construct(struct xpw_menu *xpw_menu, const char name[])
{
	xpw_construct(&xpw_menu->myself, XPW_MENU, name);
	memset(xpw_menu->entries, 0, sizeof(xpw_menu->entries));
	xpw_menu->curr_entry = 0;
	return 0;
}

void xpw_menu_clear(struct xpw_menu *xpw_menu)
{
	memset(xpw_menu->entries, 0, sizeof(xpw_menu->entries));
	xpw_menu->curr_entry = 0;
}

int xpw_menu_notifier_register(void)
{
	return xpw_class_register(&xpw_menu_notifier_class);
}

void xpw_menu_notifier_unregister(void)
{
	xpw_class_unregister(XPW_MENU_NOTIFIER);
}

int xpw_menu_register(void)
{
	int	ret;

	if((ret = xpw_class_register(&xpw_menu_class)) < 0)
		return ret;
	if((ret = xpw_hotkey_set(KEYCODE_MENU, NULL, hotkey_toplevel)) < 0)
		return ret;
	return 0;
}

void xpw_menu_unregister(void)
{
	xpw_hotkey_unset(KEYCODE_MENU);
	xpw_class_unregister(XPW_MENU);
	xpw_class_unregister(XPW_MENU_NOTIFIER);
}
