#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_passwd.h>

#define	DBG_MASK	0x80

static void dump_passwd(struct xpw *widget, int depth, const char *prefix)
{
	struct xpw_passwd	*ui_passwd;
	char			*p;
	int			i;
	int			inside = 0;

	assert(widget->classid == XPW_PASSWD);
	ui_passwd = (struct xpw_passwd *)widget;
	dump_prefix(depth, prefix);
	printf("\"%s\":\n", widget->name);
	depth++;
	dump_prefix(depth, "[");
	p = ui_passwd->contents;
	for(i = 0; i < PASSWD_MAXWIDTH && *p; i++, p++) {
		if(i == ui_passwd->cursor_pos) {
			inside = 1;
			putchar('|');
		}
		putchar(*p);
		if(i == ui_passwd->cursor_pos)
			putchar('|');
	}
	if(!inside)
		printf("|.|");
	printf("]\n");
}

int xpw_passwd_read(struct xpw_passwd *ui_passwd, char *buf, size_t buflen)
{
	return snprintf(buf, buflen, "%s", ui_passwd->contents);
}

int  xpw_passwd_set_facade(struct xpw_passwd *ui_passwd, int facade_char)
{
	ui_passwd->facade_char = facade_char;
	return 0;
}

int xpw_passwd_construct(struct xpw_passwd *ui_passwd, const char *login, const uint64_t *mask_wanted)
{
	xpw_construct(&ui_passwd->myself, XPW_PASSWD, login);
	snprintf(ui_passwd->login, sizeof(ui_passwd->login), "%s:", login);
	memset(ui_passwd->contents, 0, sizeof(ui_passwd->contents));
	ui_passwd->cursor_pos = 0;
	ui_passwd->facade_char = XPW_PASSWD_NO_FACADE;
	if (mask_wanted)
		xpw_hotkey_set_mask(WIDGET(ui_passwd), mask_wanted, NULL);
	return 0;	
}

static void redraw_passwd(struct xpw_screen *screen, struct xpw *widget)
{
	struct xpw_passwd	*ui_passwd;
	char			fake_contents[PASSWD_MAXWIDTH + 1];

	assert(widget->classid == XPW_PASSWD);
	ui_passwd = (struct xpw_passwd *)widget;
	DBG("[%s]\n", ui_passwd->contents);
	xpw_screen_gotoxy(screen, 0, 0);
	xpw_screen_string(screen, "Password:");
	xpw_screen_gotoxy(screen, 0, 1);
	if (ui_passwd->facade_char == XPW_PASSWD_NO_FACADE) {
		xpw_screen_string(screen, ui_passwd->contents);
	} else {
		char	c = ui_passwd->facade_char;
		size_t	len = strlen(ui_passwd->contents);

		memset(fake_contents, c, len);
		fake_contents[len] = '\0';
		xpw_screen_string(screen, fake_contents);
	}
	xpw_screen_gotoxy(screen, ui_passwd->cursor_pos, 1);
}

static void restart_passwd(struct xpw *widget)
{
	struct xpw_passwd	*ui_passwd;

	assert(widget->classid == XPW_PASSWD);
	ui_passwd = (struct xpw_passwd *)widget;
	DBG("[%s]\n", ui_passwd->contents);
	memset(ui_passwd->contents, 0, sizeof(ui_passwd->contents));
	ui_passwd->cursor_pos = 0;
	return;
}

static void append_char(struct xpw_passwd *ui_passwd, char ch)
{
	int	pos = ui_passwd->cursor_pos;

	if(pos < 0 || pos >= PASSWD_MAXWIDTH) {
		ERR("%s: illegal cursor position %d\n", __FUNCTION__, pos);
		return;
	}
	for(pos = 0; pos < PASSWD_MAXWIDTH && ui_passwd->contents[pos] != '\0'; pos++)
		;
	ui_passwd->contents[pos++] = ch;
	ui_passwd->contents[pos] = '\0';
	ui_passwd->cursor_pos = pos;
	DBG("%s: [%d] '%s'\n", __FUNCTION__, pos, ui_passwd->contents);
}

static void passwd_handler(struct xpw *widget, enum keycodes key)
{
	struct xpw_passwd	*ui_passwd = (struct xpw_passwd *)widget;
	struct xpw		*next_widget;

	//dump_xpw_stack("passwd_handler start");
	next_widget = NULL;
	switch(key) {
		case KEYCODE_0:
		case KEYCODE_1:
		case KEYCODE_2:
		case KEYCODE_3:
		case KEYCODE_4:
		case KEYCODE_5:
		case KEYCODE_6:
		case KEYCODE_7:
		case KEYCODE_8:
		case KEYCODE_9:
		case KEYCODE_STAR:
			if(ui_passwd->cursor_pos < PASSWD_MAXWIDTH) {
				char	ch = 0;

				if(key >= KEYCODE_0 && key <= KEYCODE_9)
					ch = '0' + (key - KEYCODE_0);
				else if(key == KEYCODE_STAR)
					ch = '*';
				if(ui_passwd->contents[ui_passwd->cursor_pos] == '\0') {
					append_char(ui_passwd, ch);
				} else {
					ui_passwd->contents[ui_passwd->cursor_pos++] = ch;
				}
			}
			break;
		case KEYCODE_DEL:
			{
				char	*p = &ui_passwd->contents[ui_passwd->cursor_pos];
				int	rest = strlen(p);

				DBG("DEL(%d)\n", rest);
				memmove(p, p+1, rest);
			}
			break;
		case KEYCODE_WEST:
			if(ui_passwd->cursor_pos > 0)
				ui_passwd->cursor_pos--;
			break;
		case KEYCODE_EAST:
			if(ui_passwd->cursor_pos < PASSWD_MAXWIDTH) {
				if(ui_passwd->contents[ui_passwd->cursor_pos] == '\0') {
					append_char(ui_passwd, ' ');
				} else {
					ui_passwd->cursor_pos++;
				}
			}
			break;
		case KEYCODE_ENTER:
			ui_passwd->contents[ui_passwd->cursor_pos] = '\0';
			ui_passwd->cursor_pos = 0;
			INFO("PASSWD(%s): Got '%s'\n", xpw_myname(WIDGET(ui_passwd)), ui_passwd->contents);
			widget_notify(widget, KEYCODE_ENTER);
			next_widget = xpw_current();
		default:
			break;
	}
	if(!next_widget)
		next_widget = widget;
	DBG("passwd_handler: \"%s\" -> \"%s\"\n", widget->name, next_widget->name);
}

static struct xpw_class	xpw_passwd_class = {
	.classid	= XPW_PASSWD,
	.name		= "Passwd",
	.kbd_handlers = {
		[KEYCODE_0]	passwd_handler,
		[KEYCODE_1]	passwd_handler,
		[KEYCODE_2]	passwd_handler,
		[KEYCODE_3]	passwd_handler,
		[KEYCODE_4]	passwd_handler,
		[KEYCODE_5]	passwd_handler,
		[KEYCODE_6]	passwd_handler,
		[KEYCODE_7]	passwd_handler,
		[KEYCODE_8]	passwd_handler,
		[KEYCODE_9]	passwd_handler,
		[KEYCODE_STAR]	passwd_handler,
		[KEYCODE_DEL]	passwd_handler,
		[KEYCODE_WEST]	passwd_handler,
		[KEYCODE_EAST]	passwd_handler,
		[KEYCODE_ENTER]	passwd_handler,
	},
	.dump_widget	= dump_passwd,
	.redraw		= redraw_passwd,
	.restart	= restart_passwd,
};

int xpw_passwd_register(void)
{
	return xpw_class_register(&xpw_passwd_class);
}

void xpw_passwd_unregister(void)
{
	xpw_class_unregister(XPW_PASSWD);
}
